# ############################################################################### #
# Autoreduction Repository : https://github.com/autoreduction/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Functionality to remove a reduction run from the database
"""
from __future__ import print_function
from typing import List, Tuple, Union

import fire
from django.db import IntegrityError
from autoreduce_scripts.manual_operations import setup_django
from autoreduce_scripts.manual_operations.util import get_run_range

setup_django()

# pylint:disable=wrong-import-position,wrong-import-order,invalid-name
from autoreduce_db.reduction_viewer.models import DataLocation, Instrument, ReductionRun, ReductionLocation


class ManualRemove:
    """
    Handles removing a run from the database
    """
    def __init__(self, instrument: str):
        """
        Args:
            instrument: The name of the instrument associated with runs
        """
        self.database = object()
        self.to_delete = {}
        self.instrument = instrument

    def find_batch_run(self, pk: int):
        """
        Finds the batch run by primary key (pk) and sets it for deletion

        Args:
            pk: The primary key of the batch run to find
        """
        # put it into list to have the same behaviour as find_run_versions_in_database
        result = [ReductionRun.objects.get(pk=pk)]
        self.to_delete[pk] = result
        return result

    def find_run_versions_in_database(self, run_number: int):
        """
        Find all run versions in the database that relate to a given instrument and run number

        Args:
            run_number: The run to search for in the database
        Returns:
            The result of the query
        """
        instrument_record, _ = Instrument.objects.get_or_create(name=self.instrument)
        result = ReductionRun.objects \
            .filter(instrument=instrument_record.id) \
            .filter(run_numbers__run_number=run_number) \
            .order_by('-created')
        self.to_delete[run_number] = list(result)
        return result

    def process_results(self, delete_all_versions: bool):
        """
        Process all the results what to do with the run based on the result of database query

        Args:
            delete_all_versions: Whether to delete all versions of the run or just the selected one
        """
        copy_to_delete = self.to_delete.copy()
        for key, value in copy_to_delete.items():
            if not value:
                self.run_not_found(run_number=key)
            if len(value) == 1:
                continue
            if len(value) > 1 and not delete_all_versions:
                self.multiple_versions_found(run_number=key)

    def run_not_found(self, run_number):
        """
        Inform user and remove key from dictionary

        Args:
            run_number: The run to remove from the dictionary
        """
        print('No runs found associated with {} for instrument {}'.format(run_number, self.instrument))
        del self.to_delete[run_number]

    def multiple_versions_found(self, run_number):
        """
        Ask the user which versions they want to remove
        Update the self.to_delete dictionary by removing unwanted versions

        Args:
            run_number: The run number with multiple versions
        """
        # Display run_number - title - version for all matching runs
        print("Discovered multiple reduction versions for {}{}:".format(self.instrument, run_number))
        for run in self.to_delete[run_number]:
            print("\tv{} - {}".format(run.run_version, run.run_description))

        # Get user input for which versions they wish to delete
        user_input = input("Which runs would you like to delete (e.g. 0,1,2,3 or 0-3): ")
        input_valid, user_input = self.validate_csv_input(user_input)
        while input_valid is False:
            user_input = input('Input was invalid. ' 'Please provide a comma separated list or a range of values: ')
            input_valid, user_input = self.validate_csv_input(user_input)

        # Remove runs that the user does NOT want to delete from the delete list
        self.to_delete[run_number] = [
            reduction_job for reduction_job in self.to_delete[run_number] if reduction_job.run_version in user_input
        ]

    def delete_records(self):
        """
        Delete all records from the database that match those found in self.to_delete
        """
        # Make a copy to ensure dict being iterated stays same size through processing
        to_delete_copy = self.to_delete.copy()
        for _, job_list in to_delete_copy.items():
            for run in job_list:
                print(f'Deleting {run.title()}')

                try:
                    run.delete()
                except IntegrityError as err:
                    print(f"Encountered integrity error: {err}\n\n"
                          "Reverting to old behaviour - manual deletion. This can take much longer.")
                    # For some reason some entries can throw an integrity error.
                    # In that case we revert to the previous (much slower) way of manually
                    # deleting everything. Perhaps there is a badly configured relation
                    # but I am not sure why it works on _most_
                    self.delete_reduction_location(run.id)
                    self.delete_data_location(run.id)
                    self.delete_reduction_run(run.id)

    @staticmethod
    def delete_reduction_location(reduction_run_id):
        """
        Delete a ReductionLocation record from the database

        Args:
            reduction_run_id: The id of the associated reduction job
        """
        ReductionLocation.objects.filter(reduction_run_id=reduction_run_id).delete()

    @staticmethod
    def delete_data_location(reduction_run_id):
        """
        Delete a DataLocation record from the database

        Args:
            reduction_run_id: The id of the associated reduction job
        """
        DataLocation.objects.filter(reduction_run_id=reduction_run_id).delete()

    @staticmethod
    def delete_reduction_run(reduction_run_id):
        """
        Delete a ReductionRun record from the database

        Args:
            reduction_run_id: The id of the associated reduction job
        """
        ReductionRun.objects.filter(id=reduction_run_id).delete()

    @staticmethod
    def validate_csv_input(user_input) -> Tuple[bool, List[int]]:
        """
        Checks if a comma separated list was provided

        Args:
            user_input: The user input to check
        Return:
            Tuple of whether the input was valid and the list of run numbers
        """
        processed_input = []
        if ',' in user_input:
            versions_to_delete = user_input.split(',')
            for number in versions_to_delete:
                try:
                    number = int(number)
                    processed_input.append(number)
                except ValueError:
                    return False, []
        elif "-" in user_input:
            range_of_versions_to_delete = user_input.split('-')
            if len(range_of_versions_to_delete) != 2:
                return False, []

            sorted_range_of_versions = sorted(map(int, range_of_versions_to_delete))
            smaller_version = int(sorted_range_of_versions[0])
            larger_version = int(sorted_range_of_versions[1])

            return True, list(range(smaller_version, larger_version + 1))
        else:
            try:
                user_input = int(user_input)
                processed_input.append(user_input)  #
            except ValueError:
                return False, []
        return True, processed_input


def remove(instrument, run_number, delete_all_versions: bool, batch_run: bool):
    """
    Run the remove script for an instrument and run_number

    Args:
        instrument: Instrument to run on
        run_number: The run number to remove
        delete_all_versions: If true, all versions of the run will be removed
        batch_run: Changes how to search for the run - normal runs are found by run number,
                   batch runs are found by their primary key, due to lack of a unique run number
    """
    manual_remove = ManualRemove(instrument)
    if not batch_run:
        manual_remove.find_run_versions_in_database(run_number)
    else:
        # parameter name is run_number but it's actually the batch run primary key!
        manual_remove.find_batch_run(run_number)
    manual_remove.process_results(delete_all_versions)
    manual_remove.delete_records()


def user_input_check(instrument, run_numbers):
    """
    User prompt for boolean value to to assert if user really wants to remove N runs

    Args:
        instrument: Instrument name related to runs user intends to remove
        run_numbers: Range of instruments submitted by user

    Returns:
        True or False to confirm removal of N runs or exit script
    """
    print(f"You are about to remove more than 10 runs from {instrument} \n"
          f"Are you sure you want to remove run numbers: {run_numbers[0]}-{run_numbers[-1]}?")
    user_input = None
    while user_input != "Y":
        user_input = input("Please enter Y: ").upper()


def main(instrument: str,
         first_run: Union[int, List[int]],
         last_run: int = None,
         delete_all_versions=False,
         no_input=False,
         batch=False):
    """
    Parse user input and run the script to remove runs for a given instrument

    Args:
        instrument: Instrument to run on
        first_run: First run to be removed.
                   If batch=True this should be the primary key of the ReductionRun object
        last_run: Optional last run to be removed
        delete_all_versions: Deletes all versions for a run without asking
        no_input: Whether to prompt the user when deleting many runs

    Returns:
        List of run numbers that were submitted.
    """
    if not isinstance(first_run, list):
        run_numbers = get_run_range(first_run, last_run=last_run)
    else:
        run_numbers = first_run

    instrument = instrument.upper()

    if not no_input and len(run_numbers) >= 10:
        user_input_check(instrument, run_numbers)

    for run in run_numbers:
        remove(instrument, run, delete_all_versions, batch)

    # ensure the range is generated when returning to the caller
    return list(run_numbers)


def fire_entrypoint():
    """
    Entrypoint into the Fire CLI interface. Used via setup.py console_scripts
    """
    fire.Fire(main)  # pragma: no cover


if __name__ == "__main__":  # pragma: no cover
    fire.Fire(main)



from enum import Enum


class RBCategory(Enum):
    DIRECT_ACCESS = "direct_access"
    RAPID_ACCESS = "rapid_access"
    COMMISSIONING = "commissioning"
    CALIBRATION = "calibration"
    INDUSTRIAL_ACCESS = "industrial_access"
    INTERNATIONAL_PARTNERS = "international_partners"
    XPESS_ACCESS = "xpess_access"
    UNCATEGORIZED = "uncategorized"



# ############################################################################### #
# Autoreduction Repository : https://github.com/autoreduction/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Settings for Nagios checks
"""
from requests.auth import HTTPBasicAuth

# ActiveMQ
ACTIVEMQ = {
    "host": "reducequeue",
    "queues": ["ReductionPending", "ReductionError"],
    "username": "YOUR-ACTIVEMQ-USERNAME",
    "password": "YOUR-ACTIVEMQ-PASSWORD",
    "api-path": ":8161/api/jolokia/read/org.apache.activemq:type=Broker,"
    "brokerName=localhost,destinationType=Queue"
}

MYSQL = {
    "host": "reducequeue",
    "username": "YOUR-MYSQL-USERNAME",
    "password": "YOUR-MYSQL-USERNAME",
    "db": "autoreduction"
}

ISIS_MOUNT = 'Z:\\'

ACTIVEMQ_URL = "http://" + ACTIVEMQ['host'] + ACTIVEMQ['api-path']
ACTIVEMQ_AUTH = HTTPBasicAuth(ACTIVEMQ['username'], ACTIVEMQ['password'])



from pathlib import Path
from unittest.mock import patch
import shutil

from django.contrib.staticfiles.testing import LiveServerTestCase
from django.utils import timezone

from autoreduce_scripts.checks import setup_django  # pylint:disable=wrong-import-order,ungrouped-imports
from autoreduce_db.reduction_viewer.models import Instrument, ReductionRun
from autoreduce_scripts.checks.daily.time_since_last_run import BASE_INSTRUMENT_LASTRUNS_TXT_DIR, main

# pylint:disable=no-member,no-self-use

setup_django()


class TimeSinceLastRunTest(LiveServerTestCase):
    """
    Test the behaviour when none of the instruments runs match the number in lastruns.txt
    """
    fixtures = ["status_fixture", "multiple_instruments_and_runs"]

    def setUp(self) -> None:
        self.instruments = Instrument.objects.all()
        for instrument in self.instruments:
            log_path = Path(BASE_INSTRUMENT_LASTRUNS_TXT_DIR.format(instrument))
            log_path.mkdir(parents=True, exist_ok=True)
            last_runs_txt = log_path / "lastrun.txt"
            last_runs_txt.write_text(f"{instrument} 44444 0")

    def tearDown(self) -> None:
        for instrument in self.instruments:
            log_path = Path(BASE_INSTRUMENT_LASTRUNS_TXT_DIR.format(instrument))
            shutil.rmtree(log_path)

    @patch("autoreduce_scripts.checks.daily.time_since_last_run.logging")
    def test_with_multiple_instruments(self, mock_logging):
        """
        Test when there are multiple instruments that haven't had run in a day.
        """
        main()
        assert mock_logging.getLogger.return_value.warning.call_count == 2

    @patch("autoreduce_scripts.checks.daily.time_since_last_run.logging")
    def test_only_one_doesnt_have_runs(self, mock_logging):
        """
        Test when one instrument hasn't had runs, but one has.
        Only one of them should cause a log message.
        """
        rr2 = ReductionRun.objects.get(pk=2)
        rr2.created = timezone.now()
        rr2.save()
        main()
        mock_logging.getLogger.return_value.warning.assert_called_once()

    @patch("autoreduce_scripts.checks.daily.time_since_last_run.logging")
    def test_all_have_runs(self, mock_logging):
        """
        Test when one instrument hasn't had runs, but one has.
        Only one of them should cause a log message.
        """
        for redrun in ReductionRun.objects.all():
            redrun.created = timezone.now()
            redrun.save()
        main()
        mock_logging.getLogger.return_value.warning.assert_not_called()

    @patch("autoreduce_scripts.checks.daily.time_since_last_run.logging")
    def test_paused_instruments_not_reported(self, mock_logging):
        """
        Test when one instrument hasn't had runs, but one has.
        Only one of them should cause a log message.
        """
        last_instr = Instrument.objects.last()
        last_instr.is_paused = True
        last_instr.save()
        main()
        mock_logging.getLogger.return_value.info.assert_called_once()
        mock_logging.getLogger.return_value.warning.assert_called_once()

    @patch("autoreduce_scripts.checks.daily.time_since_last_run.logging")
    def test_instrument_without_runs(self, mock_logging):
        """
        Test when one instrument hasn't had runs, but one has.
        Only one of them should cause a log message.
        """
        last_instr = Instrument.objects.last()
        last_instr.reduction_runs.all().delete()
        main()
        mock_logging.getLogger.return_value.warning.assert_called_once()



# ############################################################################### #
# Autoreduction Repository : https://github.com/autoreduction/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
# ! /usr/bin/env python
"""
Check the length of the queues
"""
# pylint: disable=duplicate-code
from __future__ import print_function
import sys

import requests

# pylint: disable=import-error
from autoreduce_qp.scripts.nagios_checks.autoreduce_settings import ACTIVEMQ, ACTIVEMQ_URL, ACTIVEMQ_AUTH


# pylint: disable=invalid-name
def checkQueueLength(warning, critical):
    """
    Ensures that the queue length is as expected
    :param warning: value to start reporting system warning
    :param critical: value to start reporting system critical
    :return: 0 - Success
             1 - Warning
             2 - Critical
    """
    for queue in ACTIVEMQ['queues']:
        r = requests.get(ACTIVEMQ_URL + ",destinationName=" + queue + "/QueueSize", auth=ACTIVEMQ_AUTH)

        queue_length = r.json()['value']
        # print(queue + " length = " + str(queue_length))

        if queue_length > warning:
            print(queue + " queue getting big " + str(queue_length))
            return 1
        if queue_length > critical:
            print(queue + " queue length is critical " + str(queue_length))
            return 2
    return 0


if __name__ == "__main__":
    sys.exit(checkQueueLength(3, 10))



# ############################################################################### #
# Autoreduction Repository : https://github.com/autoreduction/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
A module for creating and submitting manual submissions to autoreduction
"""
from typing import Iterable, List, Optional, Tuple, Union
import logging
import traceback

import fire
import h5py

from autoreduce_utils.clients.connection_exception import ConnectionException
from autoreduce_utils.clients.icat_client import ICATClient
from autoreduce_utils.clients.queue_client import QueueClient
from autoreduce_utils.clients.tools.isisicat_prefix_mapping import get_icat_instrument_prefix
from autoreduce_utils.message.message import Message

from autoreduce_scripts.manual_operations.rb_categories import RBCategory
from autoreduce_scripts.manual_operations import setup_django

setup_django()

# pylint:disable=wrong-import-order,wrong-import-position,no-member,too-many-arguments,too-many-return-statements

from autoreduce_db.reduction_viewer.models import ReductionRun

logger = logging.getLogger(__file__)


def submit_run(
    active_mq_client,
    rb_number: Union[str, List[str]],
    instrument: str,
    data_file_location: Union[str, List[str]],
    run_number: Union[int, Iterable[int]],
    run_title: Union[str, List[str]],
    reduction_script: str = None,
    reduction_arguments: dict = None,
    user_id=-1,
    description="",
) -> dict:
    """
    Submit a new run for autoreduction

    Args:
        active_mq_client: The client for access to ActiveMQ
        rb_number: desired experiment rb number
        instrument: name of the instrument
        data_file_location: location of the data file
        run_number: run number fo the experiment

    Returns:
        The dict representation of the message that was submitted to ActiveMQ
    """
    if active_mq_client is None:
        raise RuntimeError("ActiveMQ not connected, cannot submit runs")

    message = Message(rb_number=rb_number,
                      instrument=instrument,
                      data=data_file_location,
                      run_number=run_number,
                      facility="ISIS",
                      started_by=user_id,
                      reduction_script=reduction_script,
                      reduction_arguments=reduction_arguments,
                      description=description,
                      run_title=run_title)
    active_mq_client.send('/queue/DataReady', message, priority=1)
    logger.info("Submitted run: %s", message.serialize(indent=1))
    return message.to_dict()


def get_run_data_from_database(instrument: str, run_number: int) -> Tuple[Optional[str], Optional[str], Optional[str]]:
    """
    Retrieves a run's data-file location and rb_number from the auto-reduction database
    Args:
        database_client: Client to access auto-reduction database
        instrument: The name of the instrument associated with the run
        run_number: The run number of the data to be retrieved
    Returns:
         The data file location and rb_number, or None if this information is not in the database
    """

    # Find the latest version of a reduction run record to read information from it.
    # Does NOT include batch runs, instead looks at individual runs
    reduction_run_record = ReductionRun.objects.filter(instrument__name=instrument,
                                                       run_numbers__run_number=run_number,
                                                       batch_run=False).order_by('run_version').first()

    if not reduction_run_record:
        return None, None, None

    data_location = reduction_run_record.data_location.first().file_path
    experiment_number = str(reduction_run_record.experiment.reference_number)
    run_title = reduction_run_record.run_title

    return data_location, experiment_number, run_title


def icat_datafile_query(icat_client, file_name):
    """
    Search for file name in icat and return it if it exist.

    Args:
        icat_client: Client to access the ICAT service
        file_name: file name to search for in icat
    Returns:
        ICAT datafile entry if found
    """
    if icat_client is None:
        raise RuntimeError("ICAT not connected")

    return icat_client.execute_query("SELECT df FROM Datafile df WHERE df.name = '" + file_name +
                                     "' INCLUDE df.dataset AS ds, ds.investigation")


def get_run_data_from_icat(instrument, run_number, file_ext) -> Tuple[str, str, str]:
    """
    Retrieves a run's data-file location and rb_number from ICAT.
    Attempts first with the default file name, then with prepended zeroes.

    Args:
        icat_client: Client to access the ICAT service
        instrument: The name of instrument
        run_number: The run number to be processed
        file_ext: The expected file extension

    Returns:
        The data file location, rb_number (experiment reference) and run_title
    """
    icat_client = login_icat()

    # look for file-name assuming file-name uses prefix instrument name
    icat_instrument_prefix = get_icat_instrument_prefix(instrument)
    file_name = f"{icat_instrument_prefix}{str(run_number).zfill(5)}.{file_ext}"
    datafile = icat_datafile_query(icat_client, file_name)

    if not datafile:
        logger.info("Cannot find datafile '%s' in ICAT. Will try with zeros in front of run number.", file_name)
        file_name = f"{icat_instrument_prefix}{str(run_number).zfill(8)}.{file_ext}"
        datafile = icat_datafile_query(icat_client, file_name)

    # look for file-name assuming file-name uses full instrument name
    if not datafile:
        logger.info("Cannot find datafile '%s' in ICAT. Will try using full instrument name.", file_name)
        file_name = f"{instrument}{str(run_number).zfill(5)}.{file_ext}"
        datafile = icat_datafile_query(icat_client, file_name)

    if not datafile:
        logger.info("Cannot find datafile '%s' in ICAT. Will try with zeros in front of run number.", file_name)
        file_name = f"{instrument}{str(run_number).zfill(8)}.{file_ext}"
        datafile = icat_datafile_query(icat_client, file_name)

    if not datafile:
        raise RuntimeError(f"Cannot find datafile '{file_name}' in ICAT.")
    return datafile[0].location, datafile[0].dataset.investigation.name, datafile[0].dataset.investigation.title


def overwrite_icat_calibration_placeholder(location: str, value: Union[str, int], key: str) -> str:
    """
    Checks if the value provided has been overwritten by ICAT with calibration run placeholder text.

    Returns:
        The real value read from the NXS data file logs directly.
    """
    value = str(value)

    if "CAL" in value:
        value = read_rb_from_datafile(location, key)

    return value


def get_run_data(instrument: str, run_number: Union[str, int], file_ext: str) -> Tuple[str, str, str]:
    """
    Retrieves a run's data-file location and rb_number from the auto-reduction database,
    or ICAT (if it is not in the database)

    Args:
        instrument: The name of instrument
        run_number: The run number to be processed
        file_ext: The expected file extension

    Returns:
        The data file location and rb_number
    """
    try:
        parsed_run_number = int(run_number)
    except ValueError:
        logger.error("Cannot cast run_number as an integer. Run number given: '%s'. Exiting...", run_number)
        raise

    data_location, experiment_number, run_title = get_run_data_from_database(instrument, parsed_run_number)
    if data_location is not None and experiment_number is not None and run_title is not None:
        return data_location, experiment_number, run_title
    logger.info("Cannot find datafile for run_number %s in Auto-reduction database. "
                "Will try ICAT...", parsed_run_number)

    location, rb_num, run_title = get_run_data_from_icat(instrument, parsed_run_number, file_ext)

    # ICAT seems to do some replacements for calibration runs, overwriting the real RB number & the title
    rb_num = overwrite_icat_calibration_placeholder(location, rb_num, 'experiment_identifier')
    run_title = overwrite_icat_calibration_placeholder(location, run_title, 'title')
    return location, rb_num, run_title


def login_icat() -> ICATClient:
    """
    Log into the ICATClient

    Returns:
        The client connected, or None if failed
    """
    print("Logging into ICAT")
    icat_client = ICATClient()
    try:
        icat_client.connect()
    except ConnectionException as exc:
        print("Couldn't connect to ICAT. Continuing without ICAT connection.")
        raise RuntimeError("Unable to proceed. Unable to connect to ICAT.") from exc
    return icat_client


def login_queue() -> QueueClient:
    """
    Log into the QueueClient

    Returns:
        The client connected, or raise exception
    """
    print("Logging into ActiveMQ")
    activemq_client = QueueClient()
    try:
        activemq_client.connect()
    except (ConnectionException, ValueError) as exp:
        raise RuntimeError(
            "Cannot connect to ActiveMQ with provided credentials in credentials.ini\n"
            "Check that the ActiveMQ service is running, and the username, password and host are correct.") from exp
    return activemq_client


def windows_to_linux_path(path) -> str:
    """ Convert windows path to linux path.

    Args:
        path: The path that will be converted

    Returns:
        Linux formatted file path
    """
    # '\\isis\inst$\' maps to '/isis/'
    path = path.replace('\\\\isis\\inst$\\', '/isis/')
    path = path.replace('\\', '/')
    return path


def read_rb_from_datafile(location: str, key: str) -> str:
    """
    Reads the RB number from the location of the datafile

    Args:
        location: The location of the datafile

    Returns:
        The RB number read from the datafile
    """

    location = windows_to_linux_path(location)
    try:
        nxs_file = h5py.File(location, mode="r")
    except OSError as err:
        raise RuntimeError(f"Cannot open file '{location}'") from err

    for (_, entry) in nxs_file.items():
        try:
            return str(entry.get(key)[:][0].decode("utf-8"))
        except Exception as err:
            raise RuntimeError("Could not read RB number from datafile") from err
    raise RuntimeError(f"Datafile at {location} does not have any items that can be iterated")


def categorize_rb_number(rb_num: str):
    """
    Map RB number to a category. If an ICAT calibration RB number is provided,
    the datafile will be checked to find out the real experiment number.

    This is because ICAT will overwrite the real RB number for calibration runs!
    """
    if len(rb_num) != 7:
        return RBCategory.UNCATEGORIZED

    if rb_num[2] == "0":
        return RBCategory.DIRECT_ACCESS
    elif rb_num[2] in ["1", "2"]:
        return RBCategory.RAPID_ACCESS
    elif rb_num[2] == "3" and rb_num[3] == "0":
        return RBCategory.COMMISSIONING
    elif rb_num[2] == "3" and rb_num[3] == "5":
        return RBCategory.CALIBRATION
    elif rb_num[2] == "5":
        return RBCategory.INDUSTRIAL_ACCESS
    elif rb_num[2] == "6":
        return RBCategory.INTERNATIONAL_PARTNERS
    elif rb_num[2] == "9":
        return RBCategory.XPESS_ACCESS
    else:
        return RBCategory.UNCATEGORIZED


def main(instrument: str,
         runs: Union[int, Iterable[int]],
         reduction_script: Optional[str] = None,
         reduction_arguments: Optional[dict] = None,
         user_id=-1,
         description="") -> list:
    """
    Manually submit an instrument run from reduction.
    All run number between `first_run` and `last_run` are submitted.

    Args:
        instrument: The name of the instrument to submit a run for
        runs: The run or runs to be submitted. If a list then all the run numbers in it will be submitted
        reduction_script: The reduction script to be used. If not provided,
                          the default reduction script for the instrument will be used.
                          Currently unused as the queue processor will ignore the value
                          and always use the current reduce.py.
                          Issue tracking this https://autoreduce.atlassian.net/browse/AR-1056
        reduction_arguments: The arguments to be passed to the reduction script,
                                if None the reduce_vars.py file will be loaded
        user_id: The user ID that submitted the request. Using this script directly
                 and the run detection use -1, which is mapped to "Autoreduction service"
        description: A custom description of the run, if provided by the user

    Returns:
        A list of run numbers that were submitted.
    """

    instrument = instrument.upper()

    activemq_client = login_queue()

    submitted_runs = []

    if not isinstance(runs, Iterable):
        runs = [runs]

    for run_number in runs:
        location, rb_num, run_title = get_run_data(instrument, run_number, "nxs")
        if not location and not rb_num:
            logger.error("Unable to find RB number and location for %s%s", instrument, run_number)
            continue
        try:
            category = categorize_rb_number(rb_num)
            logger.info("Run is in category %s", category)
        except RuntimeError:
            logger.warning("Could not categorize the run due to an invalid RB number. It will be not be submitted.\n%s",
                           traceback.format_exc())
            continue

        submitted_runs.append(
            submit_run(activemq_client,
                       rb_num,
                       instrument,
                       location,
                       run_number,
                       run_title=run_title,
                       reduction_script=reduction_script,
                       reduction_arguments=reduction_arguments,
                       user_id=user_id,
                       description=description))

    return submitted_runs


def fire_entrypoint():
    """
    Entrypoint into the Fire CLI interface. Used via setup.py console_scripts
    """
    fire.Fire(main)  # pragma: no cover


if __name__ == "__main__":
    fire.Fire(main)  # pragma: no cover



# ############################################################################### #
# Autoreduction Repository : https://github.com/autoreduction/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
#! /usr/bin/env python
"""
Check the number of consumer for the list is atleast 1
"""
from __future__ import print_function
import sys

import requests

# pylint: disable=import-error
from autoreduce_qp.scripts.nagios_checks.autoreduce_settings import ACTIVEMQ, ACTIVEMQ_URL, ACTIVEMQ_AUTH


# pylint: disable=invalid-name
def checkConsumer():
    """
    checks that the consumers are consuming from queues correctly
    :return: 0 - Success
             2 - Failure
    """
    for queue in ACTIVEMQ['queues']:
        r = requests.get(ACTIVEMQ_URL + ",destinationName=" + queue + "/ConsumerCount", auth=ACTIVEMQ_AUTH)

        consumer_count = r.json()['value']
        # print(queue + " consumerCount = " + str(r.json()['value']))

        if consumer_count < 1:
            print("No consumers for  " + str(queue))
            return 2
    return 0


if __name__ == "__main__":
    sys.exit(checkConsumer())



# ############################################################################### #
# Autoreduction Repository : https://github.com/autoreduction/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
utility functions used in manual operations scripts
"""


def get_run_range(first_run, last_run=None):
    """
    Given a first_run number and optionally, a last run number return the range of runs including
    and between the run numbers. If last_run is not provided then the range will only include the
    first run
    :param first_run: (int) the first number
    :param last_run: (int) optional the second run number
    :return: (range) the range of runs including and between the first run
    """
    last_run = first_run if last_run is None else last_run
    if first_run > last_run:
        raise ValueError(f"first run: {first_run} is greater than last run: {last_run}")
    return range(first_run, last_run + 1)



# ############################################################################### #
# Autoreduction Repository : https://github.com/autoreduction/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
This script is expected to be called by a cronjob daily.

It will traverse all instrument folders in the archive (all NDX... folders),
grab reduce.py and reduce_vars.py and upload it to the repository.

The repository of the STORAGE_DIR needs to be manually configured to point to the correct
remote - otherwise this script will fail to commit/push.

There is also a --dry-run option that will just show which files are traversed,
but will not do anything.

The cronjob should be:

0 1 * * * python3 /home/reduce/backup_reduction_scripts.py

The reason it runs at 1AM is that the queue_processor restarting happens at midnight and
there might be some interference. There shouldn't be, but just in case it doesn't hurt
to be scheduled for a bit later.
"""
import os
import sys
import argparse
import datetime
import shutil
import traceback
from pathlib import Path

from git import Git, exc

# this calls the configuration of the logger which is done in the settings module
from autoreduce_utils.settings import logging

ISIS_MOUNT_PATH = Path("/isis")
AUTOREDUCTION_PATH = Path("user/scripts/autoreduction")
REDUCE_FILES_TO_SAVE = ["reduce.py", "reduce_vars.py"]

# STORAGE_DIR is the git repository dir that has been configured to point to the correct remote
STORAGE_DIR = Path("~/autoreduction_scripts").expanduser().absolute()

log = logging.getLogger(__file__)


def check_if_git_directory(path: Path):
    """
    Ensures that the directory is a git repo.

    If it is not the function will raise

    :param path: The path to the directory
    """
    repo = Git(path.absolute())
    repo.status()


def ensure_storage_exists(path: Path):
    """
    Tries to make the storage directory, if it doesn't already exist.

    If it didn't succeed raises a RuntimeError.

    :param path: The path to the directory
    """
    path.mkdir(parents=True, exist_ok=True)

    if not path.is_dir():
        raise RuntimeError(f"Could not make the '{path}' directory.")


def get_today() -> str:
    """
    Returns today in "YYYY-MM-DD" format
    """
    return str(datetime.date.today())


def commit_and_push(path: Path):
    """
    Commits all files in the path directory and pushes to origin/master

    :param path: The path to the directory
    """
    repo = Git(path.absolute())
    today = get_today()
    repo.add(".")
    repo.commit("-m", f"Reduction files for {today}")
    repo.push("--set-upstream", "origin", "master")


def main(args):
    """
    Ensures expected directories exist. Then queries the ISIS_MOUNT_PATH for all folders
    starting with NDX, and iterates through each one checking if they contain a
    reduce.py and reduce_vars.py in the expected AUTOREDUCTION_PATH.

    If the expected python files are found they are copied to STORAGE_DIR
    and later committed and pushed to the storage repository.
    """
    ensure_storage_exists(STORAGE_DIR)

    try:
        check_if_git_directory(STORAGE_DIR)
    except exc.GitCommandError as err:  # pylint: disable=no-member
        log.error(
            "Destination folder %s is not a Git repository. "
            "Please configure it manually before running this script.", str(STORAGE_DIR))
        sys.exit(1)

    for inst in [directory for directory in os.listdir(ISIS_MOUNT_PATH) if directory.startswith("NDX")]:
        path = ISIS_MOUNT_PATH / inst / AUTOREDUCTION_PATH
        destination = STORAGE_DIR / inst / AUTOREDUCTION_PATH

        ensure_storage_exists(destination)

        for file in REDUCE_FILES_TO_SAVE:
            fullpath = path / file
            log.info("Copying %s to %s", fullpath, destination)
            if not args.dry_run:
                try:
                    shutil.copy(fullpath, destination, follow_symlinks=True)
                except OSError as err:
                    log.error("Could not copy last entry. Error: %s.\n\n %s", str(err), traceback.format_exc())
    if not args.dry_run:
        commit_and_push(STORAGE_DIR)


if __name__ == '__main__':
    parser = argparse.ArgumentParser("Backup reduction scripts")
    parser.add_argument("--dry-run",
                        action="store_true",
                        help="Dry run and display all files that will be copied and to where. But do nothing!")

    _args = parser.parse_args()

    main(_args)



"""
Checks the time since the last run for each instrument.

If over 1 day it logs a warning.
"""
import logging
import os
from datetime import timedelta
from pathlib import Path

from autoreduce_utils.settings import ARCHIVE_ROOT

from django.utils import timezone

from autoreduce_scripts.checks import setup_django  # setup_django first or importing the model fails

setup_django()

# pylint:disable=wrong-import-position,wrong-import-order
from autoreduce_db.reduction_viewer.models import Instrument

# pylint:disable=no-member

BASE_INSTRUMENT_LASTRUNS_TXT_DIR = os.path.join(ARCHIVE_ROOT, "NDX{}", "Instrument", "logs")


def main():
    """
    Run through all instruments and check how long it's been since their last run.

    If the instrument is paused we don't log anything.

    The log file should then be sent to Kibana where we have alerts.
    """
    setup_django()
    logger = logging.getLogger(os.path.basename(__file__))
    instruments = Instrument.objects.all()

    for instrument in instruments:
        if instrument.is_paused:  # skip paused instruments, we are not processing runs for them
            logger.info("Instrument %s is paused", instrument)
            continue
        last_runs_txt_file = Path(BASE_INSTRUMENT_LASTRUNS_TXT_DIR.format(instrument), "lastrun.txt")
        last_runs_txt = last_runs_txt_file.read_text()

        last_run = instrument.reduction_runs.last()
        if last_run and timezone.now() - last_run.created > timedelta(1):
            # don't use batch runs to check when the last run was
            if last_run.batch_run:
                continue
            if str(last_run.run_number) not in last_runs_txt:
                logger.warning("Instrument %s has not had runs in over 1 day", instrument)
            else:
                logger.info("Last run for instrument %s matches lastrun.txt", instrument)
        else:
            logger.info("All runs OK for instrument %s", instrument)


if __name__ == "__main__":
    main()



from unittest.mock import Mock, call, patch

from django.test import TestCase

from autoreduce_scripts.manual_operations.manual_batch_submit import main as submit_batch_main
from autoreduce_scripts.manual_operations.tests.test_manual_remove import create_experiment_and_instrument


# pylint:disable=no-self-use
class TestManualBatchSubmission(TestCase):
    """
    Test manual_submission.py
    """
    fixtures = ["status_fixture"]

    def setUp(self) -> None:
        self.experiment, self.instrument = create_experiment_and_instrument()

    @patch('autoreduce_scripts.manual_operations.manual_batch_submit.login_queue')
    @patch('autoreduce_scripts.manual_operations.manual_batch_submit.submit_run')
    @patch('autoreduce_scripts.manual_operations.manual_batch_submit.get_run_data',
           return_value=("test_location", "test_rb", "test_title"))
    def test_main(self, mock_get_run_data: Mock, mock_submit_run: Mock, mock_login_queue: Mock):
        """Tests the main function of the manual batch submission"""
        runs = [12345, 12346]
        mock_reduction_script = Mock()
        mock_reduction_arguments = Mock()
        mock_user_id = Mock()
        mock_description = Mock()
        submit_batch_main(self.instrument.name, runs, mock_reduction_script, mock_reduction_arguments, mock_user_id,
                          mock_description)
        mock_login_queue.assert_called_once()

        mock_get_run_data.assert_has_calls(
            [call(self.instrument.name, runs[0], "nxs"),
             call(self.instrument.name, runs[1], "nxs")])

        mock_submit_run.assert_called_once_with(mock_login_queue.return_value,
                                                "test_rb",
                                                self.instrument.name, ["test_location", "test_location"],
                                                runs,
                                                run_title=["test_title", "test_title"],
                                                reduction_script=mock_reduction_script,
                                                reduction_arguments=mock_reduction_arguments,
                                                user_id=mock_user_id,
                                                description=mock_description)

    @patch('autoreduce_scripts.manual_operations.manual_batch_submit.login_queue')
    @patch('autoreduce_scripts.manual_operations.manual_batch_submit.submit_run')
    @patch('autoreduce_scripts.manual_operations.manual_batch_submit.get_run_data',
           side_effect=[("test_location", "test_rb", "test_title"), ("test_location_2", "test_rb_2", "test_title_2")])
    def test_main_bad_rb(self, mock_get_run_data: Mock, mock_submit_run: Mock, mock_login_queue: Mock):
        """Tests the main function of the manual batch submission"""
        runs = [12345, 12346]

        with self.assertRaises(RuntimeError):
            submit_batch_main(self.instrument.name,
                              runs,
                              reduction_script="",
                              reduction_arguments={},
                              user_id=-1,
                              description="")
        mock_login_queue.assert_called_once()

        mock_get_run_data.assert_has_calls(
            [call(self.instrument.name, runs[0], "nxs"),
             call(self.instrument.name, runs[1], "nxs")])

        mock_submit_run.assert_not_called()



# ############################################################################### #
# Autoreduction Repository : https://github.com/autoreduction/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Unit tests for manual submission utility functions
"""

from unittest import TestCase

from autoreduce_scripts.manual_operations.util import get_run_range


class TestUtil(TestCase):
    """
    Test util.py
    """
    def test_get_run_numbers_one_run_returns_correct_range(self):
        """
        Test: Correct range is returned
        When: second_run is None
        """
        self.assertEqual(range(1, 2), get_run_range(1))

    def test_get_run_numbers_valid_input_returns_valid_range(self):
        """
        Test: Correct range is returned
        When: second_run is provided
        """
        self.assertEqual(range(1, 5), get_run_range(first_run=1, last_run=4))

    def test_get_run_numbers_invalid_input_raises_value_error(self):
        """
        Test: ValueError is raised
        When: first_run is greater than last_run
        """
        with self.assertRaises(ValueError):
            get_run_range(first_run=5, last_run=2)



# ############################################################################### #
# Autoreduction Repository : https://github.com/autoreduction/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Test cases for the manual job submission script
"""
import builtins
import socket
from typing import List, Union
from unittest import mock
from unittest.mock import DEFAULT, Mock, call, patch

from autoreduce_db.reduction_viewer.models import (Experiment, Instrument, ReductionArguments, ReductionScript, Status,
                                                   DataLocation, RunNumber, ReductionRun)
from django.db import IntegrityError
from django.test import TestCase
from django.utils import timezone

from autoreduce_scripts.manual_operations.manual_remove import (ManualRemove, main, remove, user_input_check)

# pylint:disable=no-member,invalid-name


class FakeMessage:
    started_by = 0
    run_number = 1234567
    message = "I am a message"
    description = "This is a fake description"
    data = "/some/location"


def create_experiment_and_instrument():
    "Creates a test experiment and instrument"

    experiment, _ = Experiment.objects.get_or_create(reference_number=1231231)
    instrument, _ = Instrument.objects.get_or_create(name="ARMI", is_active=1, is_paused=0)
    return experiment, instrument


def _make_data_locations(reduction_run, message_data_loc: Union[str, List[str]]):
    """
    Create a new data location entry which has a foreign key linking it to the current
    reduction run. The file path itself will point to a datafile
    (e.g. "/isis/inst$/NDXWISH/Instrument/data/cycle_17_1/WISH00038774.nxs")
    """
    if isinstance(message_data_loc, str):
        DataLocation.objects.create(file_path=message_data_loc, reduction_run=reduction_run)
    else:
        DataLocation.objects.bulk_create(
            [DataLocation(file_path=data_loc, reduction_run=reduction_run) for data_loc in message_data_loc])


def _make_run_numbers(reduction_run, message_run_number: Union[int, List[int]]):
    """
    Creates the related RunNumber objects
    """
    if isinstance(message_run_number, int):
        RunNumber.objects.create(reduction_run=reduction_run, run_number=message_run_number)
    else:
        RunNumber.objects.bulk_create(
            [RunNumber(reduction_run=reduction_run, run_number=run_number) for run_number in message_run_number])


def create_reduction_run_record(experiment, instrument, message, run_version, script_text, status):
    """
    Creates an ORM record for the given reduction run and returns
    this record without saving it to the DB
    """

    time_now = timezone.now()
    script = ReductionScript.objects.create(text=script_text)
    arguments = ReductionArguments.objects.create(
        raw="""{"standard_vars":{"variable_str":"value1","variable_int":123,"variable_float":123.321,
        "variable_listint":[1,2,3],"variable_liststr":["a","b","c"],"variable_none":null,
        "variable_empty":"","variable_bool":true}}""",
        instrument=instrument)
    reduction_run = ReductionRun.objects.create(run_version=run_version,
                                                run_description=message.description,
                                                hidden_in_failviewer=0,
                                                admin_log='',
                                                reduction_log='',
                                                created=time_now,
                                                last_updated=time_now,
                                                experiment=experiment,
                                                instrument=instrument,
                                                status_id=status.id,
                                                started_by=message.started_by,
                                                reduction_host=socket.getfqdn(),
                                                batch_run=isinstance(message.run_number, list),
                                                script=script,
                                                arguments=arguments,
                                                run_title="Test title")
    _make_run_numbers(reduction_run, message.run_number)
    _make_data_locations(reduction_run, message.data)

    return reduction_run


def make_test_run(experiment, instrument, run_version: str):
    "Creates a test run and saves it to the database"
    status = Status.get_queued()
    fake_script_text = "scripttext"
    msg1 = FakeMessage()
    msg1.run_number = 101
    run = create_reduction_run_record(experiment, instrument, msg1, run_version, fake_script_text, status)
    run.save()
    run.data_location.create(file_path='test/file/path/2.raw')
    return run


def make_test_batch_run(experiment, instrument, run_version: str):
    "Creates a test run and saves it to the database"
    status = Status.get_queued()
    fake_script_text = "scripttext"
    msg1 = FakeMessage()
    msg1.run_number = [101, 102, 103]
    run = create_reduction_run_record(experiment, instrument, msg1, run_version, fake_script_text, status)
    run.save()
    run.data_location.create(file_path='test/file/path/2.raw')
    return run


class TestManualRemove(TestCase):
    """
    Test manual_remove.py
    """
    fixtures = ["status_fixture"]

    def setUp(self):
        self.manual_remove = ManualRemove(instrument="ARMI")
        # Setup database connection so it is possible to use
        # ReductionRun objects with valid meta data
        self.experiment, self.instrument = create_experiment_and_instrument()

        self.run1 = make_test_run(self.experiment, self.instrument, "1")
        self.run2 = make_test_run(self.experiment, self.instrument, "2")
        self.run3 = make_test_run(self.experiment, self.instrument, "3")

    def test_find_run(self):
        """
        Test: That the correct number of run versions are discovered
        When: find_run_versions_in_database is called
        """
        actual = self.manual_remove.find_run_versions_in_database(run_number=101)
        self.assertEqual(3, len(actual))

    def test_find_run_invalid(self):
        """
        Test: That no run versions are found for a run number that doesn't exist in the database
        When: find_run_versions_in_database is called
        """
        actual = self.manual_remove.find_run_versions_in_database(run_number="000")
        self.assertEqual(0, len(actual))

    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.run_not_found")
    def test_process_results_not_found(self, mock_not_found):
        """
        Test: run_not_found is called
        When: No matching records are found in the database
        """
        self.manual_remove.to_delete["101"] = []
        self.manual_remove.process_results(delete_all_versions=True)
        mock_not_found.assert_called_once()
        mock_not_found.reset_mock()
        self.manual_remove.process_results(delete_all_versions=True)
        mock_not_found.assert_called_once()

    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.run_not_found")
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.multiple_versions_found")
    def test_process_results_single(self, mock_multi, mock_not_found):
        """
        Test: That the code does not call multiple_versions_found
        When: The results only contain single runs (not multiple versions)
        """
        self.manual_remove.to_delete["101"] = ["test"]
        self.manual_remove.process_results(delete_all_versions=False)
        mock_multi.assert_not_called()
        mock_not_found.assert_not_called()

    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.multiple_versions_found")
    def test_process_results_multi(self, mock_multi_version):
        """
        Test: process_results function routes to correct function if the run has multiple versions
        When: Multiple runs / versions are found in the database

        Note: For this test the content of results[key] list does not have to be Run objects
        """
        self.manual_remove.to_delete["101"] = ["test", "test2"]
        self.manual_remove.process_results(delete_all_versions=False)
        mock_multi_version.assert_called_once()

    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.multiple_versions_found")
    def test_process_results_multi_with_delete_all(self, mock_multi_version):
        """
        Test: process_results function doesn't ask for user input with delete_all_versions=True
        When: Multiple runs / versions are found in the database

        Note: For this test the content of results[key] list does not have to be Run objects
        """
        self.manual_remove.to_delete["101"] = ["test", "test2"]
        self.manual_remove.process_results(delete_all_versions=True)
        mock_multi_version.assert_not_called()

    def test_run_not_found(self):
        """
        Test: That the correct corresponding run is deleted
        When: The value of a to_delete key is empty
        """
        self.manual_remove.to_delete["101"] = []
        self.manual_remove.run_not_found("101")
        self.assertEqual(0, len(self.manual_remove.to_delete))

    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.validate_csv_input")
    @patch.object(builtins, "input")
    def test_multiple_versions_found_single_input(self, mock_input, mock_validate_csv):
        """
        Test: That the user is not asked more than once for input
        When: The input is valid
        """
        self.manual_remove.to_delete["101"] = [self.run1, self.run2, self.run3]
        mock_input.return_value = "2"
        mock_validate_csv.return_value = (True, ["2"])
        self.manual_remove.multiple_versions_found("101")

        self.assertEqual(1, mock_validate_csv.call_count)
        mock_validate_csv.assert_has_calls([call("2")])

    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.validate_csv_input")
    @patch.object(builtins, "input")
    def test_multiple_versions_retry_user_input(self, mock_input, mock_validate_csv):
        """
        Test: Input is re-validated
        When: The user initially gives incorrect input
        """
        self.manual_remove.to_delete["101"] = [self.run1, self.run2, self.run3]
        mock_input.side_effect = ["invalid", "2"]
        mock_validate_csv.side_effect = [(False, []), (True, ["2"])]
        self.manual_remove.multiple_versions_found("101")

        self.assertEqual(2, mock_validate_csv.call_count)
        mock_validate_csv.assert_has_calls([call("invalid"), call("2")])

    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.validate_csv_input")
    @patch.object(builtins, "input")
    def test_multiple_versions_remove_one(self, mock_input, mock_validate_csv):
        """
        Test: That manual_remove will remove one run version
        When: Multiple versions are found for a run and one version is inputted
        """
        self.manual_remove.to_delete["101"] = [self.run1, self.run2, self.run3]
        mock_input.return_value = "2"
        mock_validate_csv.return_value = (True, ["2"])
        self.manual_remove.multiple_versions_found("101")

        # We said to delete version 2 so it should be the only entry for that run number
        self.assertEqual(1, len(self.manual_remove.to_delete["101"]))
        self.assertEqual("2", self.manual_remove.to_delete["101"][0].run_version)

    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.validate_csv_input")
    @patch.object(builtins, "input")
    def test_multiple_versions_found_list_input(self, mock_input, mock_validate_csv):
        """
        Test: The correct versions are deleted
        When: The user asks to delete two run versions as an inputted list
        """
        self.manual_remove.to_delete["101"] = [self.run1, self.run2, self.run3]
        mock_input.return_value = "1,3"
        mock_validate_csv.return_value = (True, ["1", "3"])
        self.manual_remove.multiple_versions_found("101")

        # We said to delete versions 1 and 3 so there should be those entries to delete for that run
        self.assertEqual(2, len(self.manual_remove.to_delete["101"]))
        self.assertEqual("1", self.manual_remove.to_delete["101"][0].run_version)
        self.assertEqual("3", self.manual_remove.to_delete["101"][1].run_version)

    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.validate_csv_input")
    @patch.object(builtins, "input")
    def test_multiple_versions_found_range_input(self, mock_input, mock_validate_csv):
        """
        Test: That the correct versions are deleted
        When: The user asks to delete a range of versions
        """
        self.manual_remove.to_delete["101"] = [self.run1, self.run2, self.run3]
        mock_input.return_value = "1-3"
        mock_validate_csv.return_value = (True, ["1", "2", "3"])
        self.manual_remove.multiple_versions_found("101")

        # We said to delete versions 1, 2 and 3 so there should 3 entries to delete for that run
        self.assertEqual(3, len(self.manual_remove.to_delete["101"]))
        self.assertEqual("1", self.manual_remove.to_delete["101"][0].run_version)
        self.assertEqual("2", self.manual_remove.to_delete["101"][1].run_version)
        self.assertEqual("3", self.manual_remove.to_delete["101"][2].run_version)

    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.validate_csv_input")
    @patch.object(builtins, "input")
    def test_multiple_versions_found_range_input_reversed(self, mock_input, mock_validate_csv):
        """
        Test: That the correct versions are deleted
        When: The user asks to delete a range of versions in reverse
        """
        self.manual_remove.to_delete["101"] = [self.run1, self.run2, self.run3]
        mock_input.return_value = "3-1"
        mock_validate_csv.return_value = (True, ["1", "2", "3"])
        self.manual_remove.multiple_versions_found("101")
        # We said to delete version 2 so it should be the only entry for that run number
        self.assertEqual(3, len(self.manual_remove.to_delete["101"]))
        self.assertEqual("1", self.manual_remove.to_delete["101"][0].run_version)
        self.assertEqual("2", self.manual_remove.to_delete["101"][1].run_version)
        self.assertEqual("3", self.manual_remove.to_delete["101"][2].run_version)

    def test_validate_csv_single_val(self):
        """
        Test: For expected validation result
        When: User input contains a single run
        """
        actual = self.manual_remove.validate_csv_input("1")
        self.assertEqual((True, [1]), actual)

    def test_validate_csv_single_val_invalid(self):
        """
        Test: For expected validation result (False, [])
        When: User input validation with single value that is invalid
        """
        actual = self.manual_remove.validate_csv_input("a")
        self.assertEqual((False, []), actual)

    def test_validate_csv_list(self):
        """
        Test: For expected validation result
        When: user input contains multiple runs
        """
        actual = self.manual_remove.validate_csv_input("1,2,3")
        self.assertEqual((True, [1, 2, 3]), actual)
        actual = self.manual_remove.validate_csv_input("1-3")
        self.assertEqual((True, [1, 2, 3]), actual)

        # bad input
        actual = self.manual_remove.validate_csv_input("1-2-1")
        self.assertEqual((False, []), actual)

    def test_validate_csv_invalid(self):
        """
        Test: For expected validation result (False, [])
        When: User input is invalid type and multiple
        """
        actual = self.manual_remove.validate_csv_input("t,e,s,t")
        self.assertEqual((False, []), actual)

    # pylint:disable=no-self-use
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.find_run_versions_in_database")
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.process_results")
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.delete_records")
    @patch("autoreduce_scripts.manual_operations.manual_remove.get_run_range", return_value=range(1, 2))
    def test_main(self, mock_get_run_range, mock_delete, mock_process, mock_find):
        """
        Test: The correct control functions are called
        When: The main() function is called
        """
        main(instrument="GEM", first_run=1)
        mock_get_run_range.assert_called_once_with(1, last_run=None)
        mock_find.assert_called_once_with(1)
        mock_process.assert_called_once()
        mock_delete.assert_called_once()

    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.find_run_versions_in_database")
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.process_results")
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.delete_records")
    @patch("autoreduce_scripts.manual_operations.manual_remove.get_run_range")
    def test_main_with_list(self, mock_get_run_range: Mock, mock_delete: Mock, mock_process: Mock, mock_find: Mock):
        """
        Test: The correct control functions are called
        When: The main() function is called
        """
        main(instrument="GEM", first_run=[1, 2, 3])
        mock_get_run_range.assert_not_called()
        mock_find.assert_has_calls([mock.call(1), mock.call(2), mock.call(3)])
        assert mock_process.call_count == 3
        assert mock_delete.call_count == 3

    # pylint:disable=no-self-use
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.find_run_versions_in_database")
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.process_results")
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.delete_records")
    @patch("autoreduce_scripts.manual_operations.manual_remove.user_input_check")
    @patch("autoreduce_scripts.manual_operations.manual_remove.get_run_range", return_value=range(1, 12))
    def test_main_range_greater_than_ten(self, mock_get_run_range, mock_uic, mock_delete, mock_process, mock_find):
        """
        Test: The correct control functions are called including handle_input for many runs
        When: The main() function is called
        """
        main(instrument="GEM", first_run=1, last_run=11)
        mock_get_run_range.assert_called_with(1, last_run=11)
        mock_uic.assert_called_with("GEM", range(1, 12))
        mock_find.assert_called()
        mock_process.assert_called()
        mock_delete.assert_called()

    # pylint:disable=no-self-use
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.find_run_versions_in_database")
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.process_results")
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.delete_records")
    @patch("autoreduce_scripts.manual_operations.manual_remove.get_run_range", return_value=range(1, 10))
    def test_main_range_less_than_ten(self, mock_get_run_range, mock_delete, mock_process, mock_find):
        """
        Test: The correct control functions are called including handle_input for many runs
        When: The main() function is called
        """
        return_value = main(instrument="GEM", first_run=1, last_run=9)

        assert len(return_value) == 9
        mock_get_run_range.assert_called_with(1, last_run=9)
        mock_find.assert_called()
        mock_process.assert_called()
        mock_delete.assert_called()

    # pylint:disable=no-self-use
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.find_run_versions_in_database")
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.process_results")
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.delete_records")
    def test_run(self, mock_delete, mock_process, mock_find):
        """
        Test: The correct control functions are called
        When: The run() function is called
        """
        remove("GEM", 1, False, False)
        mock_find.assert_called_once_with(1)
        mock_process.assert_called_once()
        mock_delete.assert_called_once()

    def test_delete_data_location(self):
        """
        Test: The correct query is run and associated records are removed
        When: Calling delete_data_location
        """
        with patch("autoreduce_scripts.manual_operations.manual_remove.DataLocation") as data_location:
            self.manual_remove.delete_data_location(123)
            data_location.objects.filter.assert_called_once_with(reduction_run_id=123)

    def test_delete_reduction_location(self):
        """
        Test: The correct query is run and associated records are removed
        When: Calling delete_reduction_location
        """
        with patch("autoreduce_scripts.manual_operations.manual_remove.ReductionLocation") as red_location:
            self.manual_remove.delete_reduction_location(123)
            red_location.objects.filter.assert_called_once_with(reduction_run_id=123)

    def test_delete_reduction_run_location(self):
        """
        Test: The correct query is run and associated records are removed
        When: Calling delete_reduction_run_location
        """
        with patch("autoreduce_scripts.manual_operations.manual_remove.ReductionRun") as red_run:
            self.manual_remove.delete_reduction_run(123)
            red_run.objects.filter.assert_called_once_with(id=123)

    def test_delete_records(self):
        """
        Test: Record deletion is attempted directly via the ReducedRun.delete method
        When: delete_records is called while self.to_delete is populated
        """
        mock_record = Mock()
        mock_record.id = 12
        self.manual_remove.to_delete = {"1234": [mock_record]}
        self.manual_remove.delete_records()
        mock_record.delete.assert_called_once()

    @patch.multiple("autoreduce_scripts.manual_operations.manual_remove.ManualRemove",
                    delete_reduction_location=DEFAULT,
                    delete_data_location=DEFAULT,
                    delete_reduction_run=DEFAULT)
    def test_delete_records_integrity_err_reverts_to_manual(self, delete_reduction_location, delete_data_location,
                                                            delete_reduction_run):
        """
        Test: If the ReducedRun.delete fails with Integrity Error
              the code reverts back to the manual deletion of each table entry
        When: delete_records is called while self.to_delete is populated
        """
        mock_record = Mock()
        mock_record.id = 12
        mock_record.delete.side_effect = IntegrityError
        self.manual_remove.to_delete = {"1234": [mock_record]}

        self.manual_remove.delete_records()

        delete_reduction_location.assert_called_once_with(12)
        delete_data_location.assert_called_once_with(12)
        delete_reduction_run.assert_called_once_with(12)

    def test_user_input_check(self):
        """
        Test: user_input_check() returns True of false
        When: Based on user input if range of runs to remove is larger than 10
        """
        with patch.object(builtins, "input", lambda _: "Y"):
            user_input_check(range(1, 11), "GEM")

        inputs = ["N", "Y"]
        with patch.object(builtins, "input", lambda _: inputs.pop()):
            user_input_check(range(1, 11), "GEM")


class TestManualRemoveBatchRuns(TestCase):
    """
    Test manual_remove.py
    """
    fixtures = ["status_fixture"]

    def setUp(self):
        self.manual_remove = ManualRemove(instrument="ARMI")
        # Setup database connection so it is possible to use
        # ReductionRun objects with valid meta data
        self.experiment, self.instrument = create_experiment_and_instrument()

        self.batch_run1 = make_test_batch_run(self.experiment, self.instrument, "1")
        self.batch_run2 = make_test_batch_run(self.experiment, self.instrument, "2")
        self.batch_run3 = make_test_batch_run(self.experiment, self.instrument, "3")

    # pylint:disable=no-self-use
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.find_run_versions_in_database")
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.process_results")
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.delete_records")
    def test_remove_with_mocks(self, mock_delete, mock_process, mock_find):
        """
        Test: The correct control functions are called
        When: The remove() function is called
        """
        remove("GEM", self.batch_run1.pk, False, batch_run=True)
        mock_find.assert_not_called()
        mock_process.assert_called_once()
        mock_delete.assert_called_once()

    # pylint:disable=no-self-use
    @patch("autoreduce_scripts.manual_operations.manual_remove.ManualRemove.find_run_versions_in_database")
    def test_remove_without_mocks(self, mock_find):
        """
        Test: The ReductionRun object is removed from the database
        When: The remove() function is called
        """
        pks = [self.batch_run1.pk, self.batch_run2.pk, self.batch_run3.pk]
        for pk in pks:
            remove("GEM", pk, False, batch_run=True)
            mock_find.assert_not_called()
            with self.assertRaises(ReductionRun.DoesNotExist):
                ReductionRun.objects.get(pk=pk)

    def test_find_batch_run(self):
        """
        Test: find_batch_run finds the correct run object, which should equal the one from setUp
        When: find_batch_run is called
        """
        assert self.manual_remove.find_batch_run(self.batch_run1.pk)[0] == self.batch_run1



import logging
from typing import Iterable, Optional

import fire

from autoreduce_scripts.manual_operations import setup_django

setup_django()

# pylint:disable=wrong-import-position
from autoreduce_scripts.manual_operations.manual_submission import get_run_data, login_queue, submit_run


def all_equal(iterator):
    """Tests that all elements in a list are equal"""
    iterator = iter(iterator)
    try:
        first = next(iterator)
    except StopIteration:
        return True
    return all(first == x for x in iterator)


def main(instrument,
         runs: Iterable[int],
         reduction_script: Optional[str] = None,
         reduction_arguments: Optional[dict] = None,
         user_id: int = -1,
         description: str = ""):
    """Submits the runs for this instrument as a single reduction"""

    logger = logging.getLogger(__file__)
    logger.info("Submitting runs %s for instrument %s", runs, instrument)
    instrument = instrument.upper()

    activemq_client = login_queue()
    locations, rb_numbers, titles = [], [], []
    for run in runs:
        location, rb_num, run_title = get_run_data(instrument, run, "nxs")
        locations.append(location)
        rb_numbers.append(rb_num)
        titles.append(run_title)
    if not all_equal(rb_numbers):
        raise RuntimeError("Submitted runs have mismatching RB numbers")
    return submit_run(activemq_client,
                      rb_numbers[0],
                      instrument,
                      locations,
                      runs,
                      run_title=titles,
                      reduction_script=reduction_script,
                      reduction_arguments=reduction_arguments,
                      user_id=user_id,
                      description=description)


def fire_entrypoint():
    """
    Entrypoint into the Fire CLI interface. Used via setup.py console_scripts
    """
    fire.Fire(main)  # pragma: no cover


if __name__ == "__main__":
    fire.Fire(main)  # pragma: no cover



# ############################################################################### #
# Autoreduction Repository : https://github.com/autoreduction/autoreduce
#
# Copyright &copy; 2019 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
# ! /usr/bin/env python
"""
Check that the last run is correct

Issue to convert this to use the ORM is at
https://github.com/autoreduction/autoreduce/issues/1096
"""
from __future__ import print_function
import sys
from os import path

# pylint:disable=import-error
import MySQLdb
import MySQLdb.cursors

from autoreduce_qp.scripts.nagios_checks.autoreduce_settings import MYSQL, ISIS_MOUNT


# pylint: disable=invalid-name
def checkLastRun():
    """
    Compares the last run in the database to the lastrun.txt file
    :return: 0 - Success
             2 - Failure
    """
    message = ""
    db = MySQLdb.connect(host=MYSQL['host'],
                         port=3306,
                         user=MYSQL['username'],
                         passwd=MYSQL['password'],
                         db=MYSQL['db'],
                         cursorclass=MySQLdb.cursors.DictCursor)
    cursor = db.cursor()

    # Get Instruments
    instruments = []
    cursor.execute("SELECT id," + "name FROM reduction_viewer_instrument WHERE is_active = 1 " + "AND is_paused = 0")
    for i in cursor.fetchall():
        instruments.append(i)

    # Get last reduced datafile run number
    for inst in instruments:
        cursor.execute("SELECT MAX(run_number)" + "FROM reduction_viewer_reductionrun WHERE instrument_id = " +
                       str(inst['id']))
        last_reduction_run = cursor.fetchone()['MAX(run_number)']
        with open(path.join(ISIS_MOUNT, "NDX" + inst["name"], "Instrument", "logs", "lastrun.txt")) as last_run_file:
            last_run = int(last_run_file.readline().split(" ")[1])

        # Check range because it may be a couple out.
        if last_reduction_run not in range(last_run - 2, last_run + 2):
            message += inst['name'] + " - last_run.txt = " + str(last_run) + \
                       " reduction run = " + str(last_reduction_run) + ". "

    db.close()
    if message:
        print(message)
        return 2
    return 0


if __name__ == "__main__":
    sys.exit(checkLastRun())



# ############################################################################### #
# Autoreduction Repository : https://github.com/autoreduction/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #
"""
Test cases for the manual job submission script
"""
from contextlib import contextmanager
from tempfile import NamedTemporaryFile
from unittest.mock import MagicMock, Mock, patch

import h5py
import numpy as np
from autoreduce_utils.clients.connection_exception import ConnectionException
from autoreduce_utils.clients.icat_client import ICATClient
from autoreduce_utils.clients.queue_client import QueueClient
from autoreduce_utils.message.message import Message
from django.test import TestCase
from parameterized import parameterized

from autoreduce_scripts.manual_operations import manual_submission as ms
from autoreduce_scripts.manual_operations.rb_categories import RBCategory
from autoreduce_scripts.manual_operations.tests.test_manual_remove import (FakeMessage,
                                                                           create_experiment_and_instrument,
                                                                           make_test_run)


@contextmanager
def temp_hdffile():
    """
    Writes a HDF5 file into a temporary file.
    Used in tests to ensure that experiment references are read correctly.

    The location of the experiment reference number
    in the HDF5 file is standard for ISIS NXS files.
    """
    try:
        with NamedTemporaryFile() as tmpfile:
            with h5py.File(tmpfile.name, "w") as hdffile:
                # this code writes out the string the same way it's contained
                # in the datafiles - as a numpy list of bytes
                dtype = h5py.special_dtype(vlen=bytes)
                group = hdffile.create_group("information")
                group.create_dataset("experiment_identifier", data=np.array([b"1234567"], dtype=dtype))
                group.create_dataset("title", data=np.array([b"test_title"], dtype=dtype))
                yield tmpfile
    finally:
        pass


# pylint:disable=no-self-use
class TestManualSubmission(TestCase):
    """
    Test manual_submission.py
    """
    fixtures = ["status_fixture"]

    def setUp(self):
        """ Creates test variables used throughout the test suite """
        self.valid_return = ("location", "rb", "Test title")

        self.experiment, self.instrument = create_experiment_and_instrument()

        self.run1 = make_test_run(self.experiment, self.instrument, "1")

    def mock_database_query_result(self, side_effects):
        """ Sets the return value(s) of database queries to those provided
        :param side_effects: A list of values to return from the database query (in sequence)"""
        mock_query_result = MagicMock(name="mock_query_result")
        mock_query_result.fetchall.side_effect = side_effects

    def make_query_return_object(self, return_from):
        """ Creates a MagicMock object in a format which mimics the format of
        an object returned from a query to ICAT or the auto-reduction database
        :param return_from: A string representing what type of return object
        to be mocked
        :return: The formatted MagicMock object """
        ret_obj = [MagicMock(name="Return object")]
        if return_from == "icat":
            ret_obj[0].location = self.valid_return[0]
            ret_obj[0].dataset.investigation.name = self.valid_return[1]
            ret_obj[0].dataset.investigation.title = self.valid_return[2]
        elif return_from == "db_location":
            ret_obj[0].file_path = self.valid_return[0]
        elif return_from == "db_rb":
            ret_obj[0].reference_number = self.valid_return[1]
        return ret_obj

    @patch('autoreduce_scripts.manual_operations.manual_submission.get_run_data_from_database',
           return_value=(None, None, None))
    @patch('autoreduce_scripts.manual_operations.manual_submission.get_run_data_from_icat',
           return_value=(None, None, None))
    def test_get_checks_database_then_icat(self, mock_from_icat, mock_from_database):
        """
        Test: Data for a given run is searched for in the database before calling ICAT
        When: get_run_data is called for a datafile which isn't in the database
        """
        ms.get_run_data("instrument", -1, "file_ext")
        mock_from_database.assert_called_once()
        mock_from_icat.assert_called_once()

    @patch('autoreduce_scripts.manual_operations.manual_submission.get_run_data_from_database',
           return_value=("string", 1234567, "some title"))
    @patch('autoreduce_scripts.manual_operations.manual_submission.get_run_data_from_icat',
           return_value=(None, None, None))
    def test_get_database_hit_skips_icat(self, mock_from_icat, mock_from_database):
        """
        Test: Data for a given run is searched for in the database before calling ICAT
        When: get_run_data is called for a datafile which isn't in the database
        """
        result = ms.get_run_data("instrument", -1, "file_ext")
        assert result[0] == "string"
        assert result[1] == 1234567
        assert result[2] == "some title"
        mock_from_database.assert_called_once()
        mock_from_icat.assert_not_called()

    def test_get_from_database_no_run(self):
        """
        Test: None is returned
        When: get_run_data_from_database can't find a ReductionRun record
        """
        result = ms.get_run_data_from_database('GEM', 1234567)
        for result_item in result:
            assert result_item is None

    def test_get_from_database(self):
        """
        Test: Data for a given run can be retrieved from the database in the expected format
        When: get_run_data_from_database is called and the data is present
        in the database
        """
        actual = ms.get_run_data_from_database('ARMI', 101)
        # Values from testing database
        expected = (FakeMessage().data, '1231231', 'Test title')
        self.assertEqual(expected, actual)

    @patch('autoreduce_scripts.manual_operations.manual_submission.login_icat')
    @patch('autoreduce_scripts.manual_operations.manual_submission.get_icat_instrument_prefix')
    def test_get_from_icat_when_file_exists_without_zeroes(self, _, login_icat: Mock):
        """
        Test: Data for a given run can be retrieved from ICAT in the expected format
        When: get_run_data_from_icat is called and the data is present in ICAT
        """
        login_icat.return_value.execute_query.return_value = self.make_query_return_object("icat")
        loc, rb_num, title = ms.get_run_data_from_icat("instrument", -1, "file_ext")
        login_icat.return_value.execute_query.assert_called_once()
        self.assertEqual((loc, rb_num, title), self.valid_return)

    @patch('autoreduce_scripts.manual_operations.manual_submission.login_icat')
    @patch('autoreduce_scripts.manual_operations.manual_submission.get_icat_instrument_prefix', return_value='MAR')
    def test_icat_uses_prefix_mapper(self, _, login_icat: Mock):
        """
        Test: The instrument shorthand name is used
        When: querying ICAT with function get_run_data_from_icat
        """
        icat_client = login_icat.return_value
        data_file = Mock()
        data_file.location = 'location'
        data_file.dataset.investigation.name = 'inv_name'
        data_file.dataset.investigation.title = 'inv_title'
        # Add return here to ensure we do NOT try fall through cases
        # and do NOT have to deal with multiple calls to mock
        icat_client.execute_query.return_value = [data_file]
        actual_loc, actual_inv_name, actual_inv_title = ms.get_run_data_from_icat('MARI', '123', 'nxs')
        self.assertEqual('location', actual_loc)
        self.assertEqual('inv_name', actual_inv_name)
        self.assertEqual('inv_title', actual_inv_title)
        login_icat.assert_called_once()
        icat_client.execute_query.assert_called_once_with("SELECT df FROM Datafile df WHERE"
                                                          " df.name = 'MAR00123.nxs' INCLUDE"
                                                          " df.dataset AS ds, ds.investigation")

    @patch('autoreduce_scripts.manual_operations.manual_submission.login_icat')
    @patch('autoreduce_scripts.manual_operations.manual_submission.get_icat_instrument_prefix')
    def test_get_run_data_from_icat_when_first_file_not_found(self, _, login_icat: Mock):
        """
        Test: that get_run_data_from_icat can handle a number of failed ICAT
        data file search attempts before it returns valid data file and check that
        expected format is then still returned.
        When: get_run_data_from_icat is called and the file is initially not
        found in ICAT.
        """
        # icat returns: not found a number of times before file found
        login_icat.return_value.execute_query.side_effect = [None, None, None, self.make_query_return_object("icat")]
        # call the method to test
        location_and_rb = ms.get_run_data_from_icat("instrument", -1, "file_ext")
        # how many times have icat been called
        self.assertEqual(login_icat.return_value.execute_query.call_count, 4)
        # check returned format is OK
        self.assertEqual(location_and_rb, self.valid_return)

    @patch('autoreduce_scripts.manual_operations.manual_submission.login_icat')
    @patch('autoreduce_scripts.manual_operations.manual_submission.get_icat_instrument_prefix')
    def test_get_run_data_from_icat_raises_runtimeerror(self, _, login_icat: Mock):
        """
        Test: that get_run_data_from_icat can handle a number of failed ICAT
        data file search attempts before it returns valid data file and check that
        expected format is then still returned.
        When: get_run_data_from_icat is called and the file is initially not
        found in ICAT.
        """
        # icat returns: not found a number of times before file found
        login_icat.return_value.execute_query.side_effect = [None, None, None, None]
        # call the method to test
        with self.assertRaises(RuntimeError):
            ms.get_run_data_from_icat("instrument", -1, "file_ext")

    @patch('autoreduce_scripts.manual_operations.manual_submission.get_run_data_from_database')
    @patch('autoreduce_scripts.manual_operations.manual_submission.get_run_data_from_icat')
    def test_get_when_run_number_not_int(self, mock_from_icat, mock_from_database):
        """
        Test: A SystemExit is raised and neither the database nor ICAT are checked for data
        When: get_run_data is called with a run_number which cannot be cast to int
        """
        with self.assertRaises(ValueError):
            ms.get_run_data("instrument", "invalid run number", "file_ext")
        mock_from_icat.assert_not_called()
        mock_from_database.assert_not_called()

    def test_submit_run(self):
        """
        Test: A given run is submitted to the DataReady queue
        When: submit_run is called with valid arguments
        """
        sub_run_args = {
            "active_mq_client": MagicMock(name="QueueClient"),
            "rb_number": -1,
            "instrument": "instrument",
            "data_file_location": "data_file_location",
            "run_number": -1,
            "reduction_arguments": {
                "test_int": 123,
                "test_list": [1, 2, 3]
            },
            "user_id": 15151,
            "description": "test_description",
            "run_title": "test run title"
        }
        ms.submit_run(**sub_run_args)
        message = Message(rb_number=sub_run_args["rb_number"],
                          instrument=sub_run_args["instrument"],
                          data=sub_run_args["data_file_location"],
                          run_number=sub_run_args["run_number"],
                          facility="ISIS",
                          started_by=sub_run_args["user_id"],
                          reduction_arguments=sub_run_args["reduction_arguments"],
                          description=sub_run_args["description"],
                          run_title=sub_run_args["run_title"])

        sub_run_args["active_mq_client"].send.assert_called_with('/queue/DataReady', message, priority=1)

    @patch('icat.Client')
    @patch('autoreduce_scripts.manual_operations.manual_submission.ICATClient.connect')
    def test_icat_login_valid(self, mock_connect, _):
        """
        Test: A valid ICAT client is returned
        When: We can log in via the client
        Note: We mock the connect so it does not actual perform the connect (default pass)
        """
        actual = ms.login_icat()
        self.assertIsInstance(actual, ICATClient)
        mock_connect.assert_called_once()

    @patch('icat.Client')
    @patch('autoreduce_scripts.manual_operations.manual_submission.ICATClient.connect')
    def test_icat_login_invalid(self, mock_connect, _):
        """
        Test: None is returned
        When: We are unable to log in via the icat client
        """
        con_exp = ConnectionException('icat')
        mock_connect.side_effect = con_exp
        with self.assertRaises(RuntimeError):
            ms.login_icat()
        mock_connect.assert_called_once()

    @patch('autoreduce_scripts.manual_operations.manual_submission.QueueClient.connect')
    def test_queue_login_valid(self, _):
        """
        Test: A valid Queue client is returned
        When: We can log in via the queue client
        Note: We mock the connect so it does not actual perform the connect (default pass)
        """
        actual = ms.login_queue()
        self.assertIsInstance(actual, QueueClient)

    @patch('autoreduce_scripts.manual_operations.manual_submission.QueueClient.connect')
    def test_queue_login_invalid(self, mock_connect):
        """
        Test: An exception is raised
        When: We are unable to log in via the client
        """
        con_exp = ConnectionException('Queue')
        mock_connect.side_effect = con_exp
        self.assertRaises(RuntimeError, ms.login_queue)

    def test_submit_run_no_amq(self):
        """
        Test: That there is an early return
        When: Calling submit_run with active_mq as None
        """
        with self.assertRaises(RuntimeError):
            ms.submit_run(active_mq_client=None,
                          rb_number="123",
                          instrument="instr",
                          data_file_location="loc",
                          run_number=123,
                          run_title="")

    @patch('autoreduce_scripts.manual_operations.manual_submission.login_queue')
    @patch('autoreduce_scripts.manual_operations.manual_submission.get_run_data',
           return_value=('test/file/path', "2222", "some title"))
    @patch('autoreduce_scripts.manual_operations.manual_submission.submit_run')
    def test_main_valid(self, mock_submit, mock_get_loc, mock_queue):
        """
        Test: The control methods are called in the correct order
        When: main is called and the environment (client settings, input, etc.) is valid
        """
        # Setup Mock clients
        mock_queue_client = mock_queue.return_value

        mock_reduction_script = Mock()
        mock_reduction_args = Mock()
        mock_userid = Mock()
        mock_description = Mock()

        # Call functionality
        return_value = ms.main(instrument='TEST',
                               runs=1111,
                               reduction_script=mock_reduction_script,
                               reduction_arguments=mock_reduction_args,
                               user_id=mock_userid,
                               description=mock_description)

        # Assert
        assert len(return_value) == 1
        mock_queue.assert_called_once()
        mock_get_loc.assert_called_once_with('TEST', 1111, "nxs")
        mock_submit.assert_called_once_with(mock_queue_client,
                                            "2222",
                                            'TEST',
                                            'test/file/path',
                                            1111,
                                            run_title="some title",
                                            reduction_script=mock_reduction_script,
                                            reduction_arguments=mock_reduction_args,
                                            user_id=mock_userid,
                                            description=mock_description)

    @patch('autoreduce_scripts.manual_operations.manual_submission.login_queue')
    @patch('autoreduce_scripts.manual_operations.manual_submission.get_run_data', return_value=(None, None, None))
    @patch('autoreduce_scripts.manual_operations.manual_submission.submit_run')
    def test_main_not_found_in_icat(self, mock_submit: Mock, mock_get_loc: Mock, mock_queue: Mock):
        """
        Test: The control methods are called in the correct order
        When: main is called and the environment (client settings, input, etc.) is valid
        """
        mock_reduction_script = Mock()
        mock_reduction_args = Mock()
        mock_userid = Mock()
        mock_description = Mock()

        # Call functionality
        ms.main(instrument='TEST',
                runs=1111,
                reduction_script=mock_reduction_script,
                reduction_arguments=mock_reduction_args,
                user_id=mock_userid,
                description=mock_description)

        mock_queue.assert_called_once()
        mock_get_loc.assert_called_once()
        mock_submit.assert_not_called()

    @patch('autoreduce_scripts.manual_operations.manual_submission.login_icat', side_effect=RuntimeError)
    def test_main_bad_client(self, mock_login_icat):
        """
        Test: A RuntimeError is raised
        When: Neither ICAT or Database connections can be established
        """
        self.assertRaises(RuntimeError, ms.main, 'TEST', 1111)
        mock_login_icat.assert_called_once()

    @parameterized.expand([
        ["210NNNN", RBCategory.DIRECT_ACCESS],
        ["211NNNN", RBCategory.RAPID_ACCESS],
        ["212NNNN", RBCategory.RAPID_ACCESS],
        ["2130NNN", RBCategory.COMMISSIONING],
        ["2135NNN", RBCategory.CALIBRATION],
        ["215NNNN", RBCategory.INDUSTRIAL_ACCESS],
        ["216NNNN", RBCategory.INTERNATIONAL_PARTNERS],
        ["219NNNN", RBCategory.XPESS_ACCESS],
        ["214NNNN", RBCategory.UNCATEGORIZED],
        ["217NNNN", RBCategory.UNCATEGORIZED],
        ["218NNNN", RBCategory.UNCATEGORIZED],
        ["99999999999", RBCategory.UNCATEGORIZED],
        ["0", RBCategory.UNCATEGORIZED],
        ["1234", RBCategory.UNCATEGORIZED],
    ])
    def test_categorize_rb_number(self, rb_num, expected_category):
        """
        Test that the RB number is categorized correctly depending on the numbers in it.
        """
        assert ms.categorize_rb_number(rb_num) == expected_category

    def test_read_rb_from_datafile(self):
        """
        Test that the rb number can be read from the datafile's
        experiment_reference HDF5 group
        """
        with temp_hdffile() as tmpfile:
            # replace / with \\ so that it looks like a Windows path
            ms.read_rb_from_datafile(tmpfile.name.replace("/", "\\\\"), "experiment_identifier")

    def test_read_rb_from_datafile_no_rb_number(self):
        """
        Test that a RuntimeError is raised if the rb number is
        not found in the datafile's experiment_reference HDF5 group
        """
        with NamedTemporaryFile() as tmpfile:
            with h5py.File(tmpfile.name, "w") as hdffile:
                hdffile.create_group("information")

            # replace / with \\ so that it looks like a Windows path
            with self.assertRaises(RuntimeError):
                ms.read_rb_from_datafile(tmpfile.name.replace("/", "\\\\"), "experiment_identifier")

    def test_read_rb_from_datafile_empty_nxs(self):
        """
        Test that a RuntimeError is raised if the file provided is empty
        """
        with NamedTemporaryFile() as tmpfile:
            with h5py.File(tmpfile.name, "w"):
                pass

            # replace / with \\ so that it looks like a Windows path
            with self.assertRaises(RuntimeError):
                ms.read_rb_from_datafile(tmpfile.name.replace("/", "\\\\"), "experiment_identifier")

    def test_icat_datafile_query(self):
        """
        Test that RuntimeError is raised if the icat_client provided is None
        """
        with self.assertRaises(RuntimeError):
            ms.icat_datafile_query(None, "test")

    def test_overwrite_icat_calibration_placeholder(self):
        """
        Test that runs with CAL_<some string> get their RB overwritten with
        one from the datafile
        """
        with temp_hdffile() as tmpfile:
            assert ms.overwrite_icat_calibration_placeholder(tmpfile.name, "CAL_TEST",
                                                             "experiment_identifier") == "1234567"

            assert ms.overwrite_icat_calibration_placeholder(tmpfile.name, "CAL_TEST", "title") == "test_title"

    def test_no_overwrite_icat_calibration_placeholder(self):
        """
        Test that the RB is not overwritten when CAL is not present in the string
        """
        with temp_hdffile() as tmpfile:
            assert ms.overwrite_icat_calibration_placeholder(tmpfile.name, "test_rb_number",
                                                             "experiment_identifier") == "test_rb_number"
            assert ms.overwrite_icat_calibration_placeholder(tmpfile.name, "some test title",
                                                             "title") == "some test title"



# pylint:skip-file
"""
Wrapper for the functionality for various installation and project setup commands
see:
    `python setup.py help`
for more details
"""
from setuptools import setup, find_packages

setup(
    name="autoreduce_scripts",
    version="22.0.0.dev28",
    description="ISIS Autoreduce helper scripts",
    author="ISIS Autoreduction Team",
    url="https://github.com/autoreduction/autoreduce-scripts/",
    install_requires=[
        "autoreduce_db==22.0.0.dev18",
        "autoreduce_utils==22.0.0.dev7",
        "django==3.2.8",
        "fire==0.4.0",
        "h5py==2.10.0",  # for reading the RB number from the datafile
        "GitPython==3.1.14"  # for backup_reduction_scripts.py
    ],
    packages=find_packages(),
    entry_points={
        "console_scripts": [
            "autoreduce-manual-submission = autoreduce_scripts.manual_operations.manual_submission:fire_entrypoint",
            "autoreduce-manual-remove = autoreduce_scripts.manual_operations.manual_remove:fire_entrypoint",
            "autoreduce-check-time-since-last-run = autoreduce_scripts.checks.daily.time_since_last_run:main"
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3 :: Only",
    ])



"""
Django settings for autoreduce_django project.

Generated by 'django-admin startproject' using Django 3.2.2.

For more information on this file, see
https://docs.djangoproject.com/en/3.2/topics/settings/

For the full list of settings and their values, see
https://docs.djangoproject.com/en/3.2/ref/settings/
"""
from pathlib import Path

from autoreduce_db.autoreduce_django.settings import DATABASES as autoreduce_db_settings

# Build paths inside the project like this: BASE_DIR / 'subdir'.
BASE_DIR = Path(__file__).resolve().parent.parent

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/3.2/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = 'django-insecure-1_r@nia!d7&9q)yigp=+p5i)p@ei+b(#w71z4ap%)r$9g6-9i!'

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []

# Application definition
INSTALLED_APPS = [
    'autoreduce_scripts.autoreduce_django',
    'autoreduce_db.reduction_viewer',
    'autoreduce_db.instrument',
]
DATABASES = autoreduce_db_settings

MIDDLEWARE = []

ROOT_URLCONF = 'autoreduce_django.urls'

TEMPLATES = []

WSGI_APPLICATION = 'autoreduce_django.wsgi.application'

# Password validation
# https://docs.djangoproject.com/en/3.2/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = []

# Internationalization
# https://docs.djangoproject.com/en/3.2/topics/i18n/

LANGUAGE_CODE = 'en-us'

TIME_ZONE = 'UTC'

USE_I18N = True

USE_L10N = True

USE_TZ = True

# Default primary key field type
# https://docs.djangoproject.com/en/3.2/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = 'django.db.models.BigAutoField'



from pathlib import Path
from unittest.mock import patch
import shutil
from django.contrib.staticfiles.testing import LiveServerTestCase

from autoreduce_scripts.checks import setup_django  # pylint:disable=wrong-import-order,ungrouped-imports
from autoreduce_db.reduction_viewer.models import Instrument
from autoreduce_scripts.checks.daily.time_since_last_run import BASE_INSTRUMENT_LASTRUNS_TXT_DIR, main

# pylint:disable=no-member,no-self-use

setup_django()


class TimeSinceLastRunMatchingLastrunsTest(LiveServerTestCase):
    """
    Test the behaviour when the last runs matches the lastruns.txt contents
    """
    fixtures = ["status_fixture", "multiple_instruments_and_runs"]

    def setUp(self) -> None:
        self.instruments = Instrument.objects.all()
        for instrument in self.instruments:
            log_path = Path(BASE_INSTRUMENT_LASTRUNS_TXT_DIR.format(instrument))
            log_path.mkdir(parents=True, exist_ok=True)
            last_runs_txt = log_path / "lastrun.txt"
            last_runs_txt.write_text(f"{instrument} {instrument.reduction_runs.last().run_number} 0")

    def tearDown(self) -> None:
        for instrument in self.instruments:
            log_path = Path(BASE_INSTRUMENT_LASTRUNS_TXT_DIR.format(instrument))
            shutil.rmtree(log_path)

    @patch("autoreduce_scripts.checks.daily.time_since_last_run.logging")
    def test_with_multiple_instruments(self, mock_logging):
        """
        Test when there are multiple instruments that haven't had run in a day, but they are also
        the last runs recorded in lastruns.txt - we are not expecting anything to be logged, as the
        beamline hasn't had any new runs that need processing
        """
        main()
        mock_logging.getLogger.return_value.warning.assert_not_called()
        assert mock_logging.getLogger.return_value.info.call_count == 2



# ############################################################################### #
# Autoreduction Repository : https://github.com/autoreduction/autoreduce
#
# Copyright &copy; 2020 ISIS Rutherford Appleton Laboratory UKRI
# SPDX - License - Identifier: GPL-3.0-or-later
# ############################################################################### #

import django
from django.conf import settings
from autoreduce_scripts.autoreduce_django.settings import DATABASES, INSTALLED_APPS


def setup_django():
    """
    Sets up django if not configured already. This allows accessing the models through the ORM
    """

    if not settings.configured:
        settings.configure(DATABASES=DATABASES, INSTALLED_APPS=INSTALLED_APPS, USE_TZ=True)
        django.setup()


setup_django()



#!/usr/bin/env python
"""Django's command-line utility for administrative tasks."""
import os
import sys


def main():
    """Run administrative tasks."""
    os.environ.setdefault('DJANGO_SETTINGS_MODULE', 'autoreduce_django.settings')
    try:
        # pylint:disable=import-outside-toplevel
        from django.core.management import execute_from_command_line
    except ImportError as exc:
        raise ImportError("Couldn't import Django. Are you sure it's installed and "
                          "available on your PYTHONPATH environment variable? Did you "
                          "forget to activate a virtual environment?") from exc
    execute_from_command_line(sys.argv)


if __name__ == '__main__':
    main()
